# $Id: uploadthread.py,v 1.19 2010/11/22 17:44:57 irees Exp $
import os
import operator
import Queue
import collections
import time
import datetime
import json

from PyQt4 import QtCore

import emen2.clients.emdash.config
import emen2.clients.handlers


class UploadThread(QtCore.QThread):
	"""Upload a file"""
	
	signal_file_failure = QtCore.pyqtSignal(unicode, unicode)
	signal_file_status = QtCore.pyqtSignal(unicode, object)
	signal_file_success = QtCore.pyqtSignal(unicode, object)	
	
	def __init__(self, parent=None, queue=None, db=None):
		self.db = db
		self.queue = queue

		self.options = {
			"sidecar":True,
			"retry":False,
			"quiet":True
		}
		
		QtCore.QThread.__init__(self, parent=parent)
		


	def enqueue(self, filename, item):
		item["_status"] = "Queued"
		self.queue.put(item)
		self.signal_file_status.emit(filename, item)



	
	def report(self, msg=None, progress=None, f=None, pos=None, i=0):
		if progress == None:
			progress = 0.0
		
		if f:		
			self.signal_file_status.emit(f, {"_status":msg or progress, "_progress": progress})
		
		
	
	def run(self):		
		while True:
			try:
				self.upload()
			except Exception, e:
				print e
				break


						
	def upload(self):					
		
		item = self.queue.get()
		filename = item.get('_filename')
		recid = item.get('_parent')
		

		# Note that _filename is used as either an actual filename, or an identifier
		# This causes the thread to terminate
		if filename == None:
			self.queue.task_done()
			raise Exception, "No more files"

		if recid < 0:
			raise Exception, "UploadThread: Invalid recid for upload: %s"%recid

		# wait a small amount of time before proceeding..
		time.sleep(emen2.clients.emdash.config.sleeptime)		

		# Remove the "meta-parameters" used for queue details
		rec = {}
		for k,v in item.items():
			if not k.startswith("_"):
				rec[k] = v												
				
		try:
			handler = emen2.clients.handlers.gethandler(rectype=rec.get('rectype'), filename=filename)
			dbt = handler(db=self.db, filename=filename, recid=recid, options=self.options, applyparams=rec)
			dbt.addlistener(self.report)
			dbt.upload()

			# This is not very nice...
			# rec = ret.get(filename, dict()).get('recid')
			rec = dbt.get_upload_target()
			if rec:
				rec = self.db.getrecord(rec)
				self.signal_file_success.emit(filename, rec)
		
		
		except Exception, e:
			print "UploadThread exception: %s"%e
			self.signal_file_failure.emit(filename, unicode(e))
			self.queue.put(item)
			time.sleep(5)


		self.queue.task_done()
	
	
__version__ = "$Revision: 1.19 $".split(":")[1][:-1].strip()


