# $Id: proxy.py,v 1.17 2011/03/07 09:45:30 irees Exp $
import copy
import urllib
import urlparse
import itertools
import xmlrpclib
import random
import time
import UserDict, collections
collections.Mapping.register(UserDict.DictMixin)

from hashlib import sha1

import jsonrpc.proxy
import emen2.clients
import emen2.util.jsonutil
import emen2.db.paramdef
import emen2.db.recorddef
import emen2.db.record
import emen2.db.user




##################################
# Moved from clients/emen2proxy
##################################


def zipdicts(*dicts):
	keydict = dicts[0]
	for k in keydict.iterkeys():
		yield (k, tuple(dict_.get(k) for dict_ in dicts))



def newrecord(rec):
	r = emen2.db.record.Record(rec)
	return r



def newthing(typ):
	def _inner(data):
		a = object.__new__(typ)
		a.update(data)
		return a
	return _inner




class Context(object):
	def __init__(self, ctxid, host):
		self.ctxid = ctxid
		self.host = host
		self.db = None






##################################
# Main exported DB interface classes
##################################



# JSONRPCProxy customized for EMEN2

class EMEN2JSONRPCProxy(jsonrpc.proxy.JSONRPCProxy):
	def init(self, ctxid=None, t=False, *args, **kwargs):
		self._ctx = Context(ctxid, self.serviceURL)
		self._host = self.serviceURL
		self._ctx.db = self
		self._convert = True
		self._t = t

	def transformURL(self, url, path):
		if self._t:
			path = '/?'.join([path, 't=1'])
		return url, path

	typemapping = dict(
		set=set,
		tuple=tuple,
		record=newrecord,
		paramdef=newthing(emen2.db.paramdef.ParamDef),
		recorddef=newthing(emen2.db.recorddef.RecordDef),
		user=newthing(emen2.db.user.User)
	)
	revmap = dict( (y,x) for x,y in typemapping.iteritems() )

	def restore_from_typemap(self, typemap, val):
		result = val
		containertype = None
		if isinstance(typemap, list):
			result = {}
			for k, (vv, tv) in zipdicts(val, typemap[1]):
				result[k] = self.restore_from_typemap(tv, vv)
			containertype=self.typemapping.get(typemap[0])
		else:
			containertype=self.typemapping.get(typemap)
		if containertype is not None:
			result = containertype(result)
		return result

	def __getattr__(self, name):
		result = jsonrpc.proxy.JSONRPCProxy.__getattr__(self, name)
		result.set_ctxid(self._ctx.ctxid)
		return result

	def get_ctxid(self):
		return self._ctx.ctxid

	def set_ctxid(self, ctxid):
		self._ctx.ctxid = ctxid

	def get_postdata(self, args, kwargs):
		if self._ctx.ctxid: kwargs['ctxid'] = self._ctx.ctxid
		return jsonrpc.proxy.JSONRPCProxy.get_postdata(self, args, kwargs)

	def proc_response(self, data):
		result = copy.deepcopy(data['result'])
		typemap = copy.deepcopy(data.get('type'))
		if typemap:
			result = self.restore_from_typemap(typemap, result)
		result = dict(
			result = result
		)
		return result

	def _login(self, username, password):
		self._ctx.ctxid = self.login(username, password)




# Old-style XML-RPC Proxy. Does not support keyword arguments on method calls!!

class EMEN2XMLRPCProxy(xmlrpclib.ServerProxy):
	"""Provide a more feature-rich DB XMLRPC interface than the bare python xmlrpclib"""

	def __init__(self, host, username=None, password=None, ctxid=None):
		xmlrpclib.ServerProxy.__init__(self, host, allow_none=True)

		self._username = username
		self._password = password

		self._ServerProxy__allow_none = True
		self._host = host
		self._handler = self._ServerProxy__handler

		self._setContext(ctxid)


	def _wrap_request(self, method, params):
		try:
			ret = self._ServerProxy__request(method, params)
		except xmlrpclib.Error, err:
			#raise xmlrpclib.Fault(0, "Error: %s"%(err.headers.get('X-Error')))
			raise ValueError, str(err.headers.get('X-Error'))

		except Exception, e:
			raise

		return ret


	def __getattr__(self, name):
		# print "-> ", name
		return xmlrpclib._Method(self._wrap_request, name)


	def _login(self, username=None, password=None):
		ctxid = self._wrap_request("login", (username or self._username, password or self._password))
		self._password = None
		return self._setContext(ctxid)


	def _setContext(self, ctxid):
		# ian: todo: make handler a dynamically generated property? maybe not..
		self._ctxid = ctxid
		if self._ctxid != None:
			self._ServerProxy__handler = "%s?ctxid=%s"%(self._handler, self._ctxid)
		return self._ctxid


	def _getContext():
		return self._ctxid





if __name__ == "__main__":
	db = emen2.clients.opendb()


__version__ = "$Revision: 1.17 $".split(":")[1][:-1].strip()

