# $Id: publicresource.py,v 1.57 2011/03/07 17:06:33 edwlan Exp $
import urllib2
import emen2.util.jsonutil
import re
import time
import collections


import twisted.web

# emen2 imports
import emen2.web.resources.emen2resource
import emen2.db.exceptions
from emen2.web import routing
from emen2.web import responsecodes
#from .authresource import render_security_error

import emen2.db.config
g = emen2.db.config.g()




def force_unicode(item):
	'''force objects to abide by python's absurd str/unicode problems and Twisted's equally broken handling of them'''
	result = item
	if isinstance(item, str): pass
	elif isinstance(item, unicode):
		result = item.encode('utf-8')
	elif hasattr(item, '__iter__'):
		if hasattr(item, 'items'):
			result = dict()
			for key, value in item.items():
				if value == None: result[force_unicode(key)] = value
				else: result[force_unicode(key)] = force_unicode(value)
		else:
			typ = type(item)
			result = typ(force_unicode(x) for x in item)
	else:
		result = str(item)
	return result





class PublicView(emen2.web.resources.emen2resource.EMEN2Resource):

	resourcename = 'public'
	router = routing.URLRegistry()
	redirects = {}
	openrequests = set()
	isLeaf = False

	##### Resource overrides #####
	def getChildWithDefault(self, path, request):
		d = request.notifyFinish()
		if not id(request) in self.openrequests:
			# Cache the id of the request so that we can keep track of the requests that have started
			self.openrequests.add(id(request))

			# save time so that we can get a request length in logrequest
			# (it is more accurate to do it here than elsewhere, since there is less latency between
			#  network activity and entering/exiting this method)
			t = time.time()

			# register the logging callback
			# done here so that we can log access to static files as well as views
			d.addCallback(self.logrequest, request, t=t)

			# pop the current request from openrequests
			d.addCallback(self.poprequest, request)

			# make sure errors in these callbacks get handled
			d.addErrback(g.log_error)

		# copied functionality of Resource.getChild
		result = self.getChild(path, request)
		return result


	def getChild(self, path, request):
		result = self
		if request.prepath[0] in self.children:
			result = self.children.get(request.prepath[0])
		return result


	def poprequest(self, _, request):
		assert id(request) in self.openrequests
		self.openrequests.remove(id(request))


	def logrequest(self, x, request, t=0.0, *args, **kwargs):
		ctxid = None
		headers = request.responseHeaders.getAllRawHeaders()
		tmp = {}
		for k, v in headers:
			k = k.lower()
			if k not in tmp and k in set(['x-username', 'content-length', 'x-resource']):
				tmp[k] = v

		logline = emen2.util.loganalyzer.AccessLogLine(
			host = request.getClientIP(),
			ctxid = request.getCookie('ctxid'),
			username = tmp.get('x-username',[''])[0],
			rtime = time.strftime(emen2.util.loganalyzer.CTIME),
			request = '"%s %s"' % (request.method, request.uri),
			response_code = request.code,
			size = tmp.get('content-length',[0])[0],
			resource = tmp.get('x-resource', '-'),
			cputime = (time.time()-t)*1000,
		)
		g.log.msg('LOG_WEB', str(logline))



	# public resource actions
	@classmethod
	def getredirect(cls, name):
		# g.debug(cls.redirects)
		redir = cls.redirects.get(name, False)
		result = None
		if redir != False:
			to, args, kwargs = redir
			g.debug(to, args, kwargs)
			result = routing.URLRegistry.reverselookup(to, *args, **kwargs)
		return result


	@classmethod
	def register_redirect(cls, fro, to, *args, **kwargs):
		cls.redirects[fro] = (to, args, kwargs)


	def parse_uri(self, uri):
		base, _, qs = uri.partition('?')
		return base, qs


	def _getredirect(self, target, request, path):
		npath = None
		base, qs = self.parse_uri(request.uri)

		#registered redirect
		redir = self.getredirect(npath or path)
		if redir != None: npath = redir

		#redirect if necessary
		if npath != None:
			if qs: qs = str.join('', ('?', qs))
			res = ''.join((npath, qs))
			if g.EMEN2WEBROOT:
				res = '/'.join((g.EMEN2WEBROOT,str.join('/', request.prepath)))
			target = '%s/%s%s%s' % (g.EMEN2WEBROOT,str.join('/', request.prepath), npath, qs)

		#return new target or None for no redirect
		return target


	def _action(self, request=None, args=None, ctxid=None, host=None, db=None, **_):
		'''set db context, call view, and get string result
		put together to minimize amount of blocking code'''

		ret = None

		g.debug(__name__, ":::", args, request.args)
		# Get path; don't use postpath because now isLeaf=True, so it will be empty
		# This Resource used to be /db; don't break people's bookmarks
		path = request.path
		path = path.replace("/db/","/")

		# Lack of trailing slash used to cause redirect; I think I'll just add it "secretly"
		if not path:
			path = '/'
		if path[-1] != "/":
			path = "%s/"%path

		# don't remove this, _getredirect does other things besides adding the slash
		target = self._getredirect(None, request, path)
		# Redirect if necessary
		if target is not None:
			g.log.msg('LOG_INFO', 'redirected %r to %r' % (request.uri, target))
			raise responsecodes.HTTPMovedPermanently('', target)

		# ian: experiment.. if ctxid is passed as a qs arg, it will set the cookie.
		setctxid = request.args.pop('ctxid', None)

		# completely stupid handling of args to make sure they are str, not unicode..
		args2 = {}
		if args:
			for k,v in args.items():
				args2[str(k)] = v
			args = args2

		callback = routing.URLRegistry().execute(path, method=request.method, fallback='GET', **args)

		# this binds the Context to the DBProxy for the duration of the view
		username = 'anonymous'
		with db:
			db._setContext(ctxid,host)
			result = callback(db=db, method=request.method)
			headers = result.get_headers()

			if hasattr(result, 'register_as_producer'):
				# allow the view to register a twisted producer
				result.register_as_producer(request)

			else:
				ret = ''

				# etags allow us to see if we need to actually resend the page data
				# ---
				# if the result defines an etag, try to set it in the request, the return
				# value will indicate whether or not we need to render the page
				etag = None
				if result.etag != None:
					etag = request.setETag(result.etag)

				if result.etag == None or etag != twisted.web.http.CACHED:
					ret = result.get_data()

				username = db._getctx().username

				# Neither of these are actually used, although they might be useful for one thing or another
				headers['X-Username'] = username
				if setctxid: headers['X-ctxid'] = setctxid

				# This is used for logging
				headers['X-Resource'] = 'public'

		# ret and force_unicode(ret) will return None if ret is None
		return ret and force_unicode(ret), force_unicode(headers)


__version__ = "$Revision: 1.57 $".split(":")[1][:-1].strip()
