# $Id: rpcresource.py,v 1.39 2011/03/05 00:11:37 irees Exp $
import os
import time
import xmlrpclib
import emen2.util.jsonutil


# emen2 imports
import emen2.db.config
g = emen2.db.config.g()
import emen2.util.loganalyzer
import emen2.util.listops as listops
import emen2.web.resources.emen2resource


#import emen2.db

Fault = xmlrpclib.Fault




class RPCFormatJSON:
	def __init__(self):
		pass

	def decode(self, content, kw):
		args = emen2.util.jsonutil.decode(content)
		kwargs = {}
		if hasattr(args,"items"):
			for k,v in args.items():
				kwargs[str(k)]=v
			args=[]

		return None, tuple(args), kwargs

	def encode(self, method, value):
		value = emen2.util.jsonutil.encode(value, escape_unicode=True)
		return value.encode("utf-8", 'replace')




class RPCFormatXMLRPC:
	def __init__(self):
		pass

	def decode(self, content, kw):
		args, method = xmlrpclib.loads(content)

		args=list(args)
		kwargs = {}
		# add kwargs support to xmlrpc:: if last arg is dict...
		#if isinstance(args[-1],dict):
		#	kwargs = args.pop()

		# not really supported... no decode; values limited to strings.
		for k,v in kw.items():
			try:
				kwargs[str(k)] = v[0]
			except:
				pass

		if hasattr(args,"items"):
			for k,v in args.items():
				kwargs[str(k)]=v
			args=[]

		return method, args, kwargs


	def encode(self, method, value):
		value = self.encode_serialize(value)

		if not isinstance(value, tuple):
			value = (value,)
		value = xmlrpclib.dumps(value, methodresponse=1, allow_none=True)

		return value


	def encode_serialize(self, value):
		"""Serializes in UTF-8 DB instances for unsophisticated encoders (e.g. xmlrpc)"""
		# convert to dict using class method
		if isinstance(value, (emen2.db.dataobject.BaseDBObject)):
			return self.encode_serialize(dict(value))

		elif hasattr(value,"items"):
			v2={}
			for k,v in value.items():
				v2[str(k)]=self.encode_serialize(v)
			return v2

		elif hasattr(value,"__iter__"):
			ret=[]
			for i in value:
				ret.append(self.encode_serialize(i))
			return ret

		elif type(value)==unicode:
			value=value.encode("utf-8")

		return value




class RPCResource(emen2.web.resources.emen2resource.EMEN2Resource):

	isLeaf = True
	resourcename = 'rpc'

	def __init__(self, format="xmlrpc"):
		emen2.web.resources.emen2resource.EMEN2Resource.__init__(self)

		self.format = format
		if format=="json":
			self.handler = RPCFormatJSON()
			self.fmt = 'application/json'
		elif format=="xmlrpc":
			self.handler = RPCFormatXMLRPC()
			self.fmt = 'text/xml'


	def parse_content(self, request):
		postargs = {}
		request.content.seek(0, 0)
		content = request.content.read()
		method, args, kwargs = self.handler.decode(content, request.args)
		postargs['method_call'] = method
		postargs['method_args'] = args
		postargs['method_kwargs'] = kwargs
		return postargs



	def _action(self, args=None, ctxid=None, host=None, db=None, request=None):
		# this binds the Context to the DBProxy for the duration of the view

		method_call = args.get('method_call')
		method_args = args.get('method_args')
		method_kwargs = args.get('method_kwargs')

		ctxid = ctxid or method_kwargs.pop("ctxid",None)

		if method_call == None:
			method_call = '.'.join(request.postpath)

		# this is a compatibility hack for old-style getchildren/getparents and getbinary
		# this may be removed at some point in the future...
		# ian: todo: remove
		if method_call in ["getchildren","getparents"]:
			try:
				# print "getparents/getchildren compat fix"
				if args[1] == "record":
					args.pop(1)
				args = args[:3]
			except:
				pass


		if method_call == "login" or method_call == "_login":
			method_kwargs['host'] = host


		db._starttxn(write=db._checkwrite(method_call))
		with db:
			db._setContext(ctxid, host)
			username = db._getctx().username
			result = db._callmethod(method_call, method_args, method_kwargs)


		# old-style getbinary wants a dictionary...
		# like the above fix, this will almost certainly be removed later
		# ian: todo: remove
		if method_call == "getbinary":
			try:
				if args[1] == True:
					result = listops.dictbykey(result, 'name')
			except:
				pass


		return self.handler.encode(method_call, result), {'X-Username': username,
																		"Content-Type": self.fmt,
																		'X-Resource': 'rpc'}



__version__ = "$Revision: 1.39 $".split(":")[1][:-1].strip()
