# $Id: auth.py,v 1.30 2011/03/05 00:11:37 irees Exp $
import urlparse

# Standard View imports
import emen2.db.config
g = emen2.db.config.g()
from emen2.web.view import View
###

@View.register
class Auth(View):

	@View.add_matcher(r'^/auth/login/$')
	def login(self, username=None, pw=None, msg='', errmsg='', location=None, **kwargs):
		self.template = '/pages/auth.login'
		self.title = 'Login'
		location = location or g.EMEN2WEBROOT
		if 'auth' in location or not location:
			location = g.EMEN2WEBROOT

		self.set_context_item("username",username)
		self.set_context_item('location', location)
		ctxid = None
		
		if username != None:
			#try:
			ctxid = self.db.login(username, pw, host=self.ctxt['HOST'])
			msg = 'Successfully logged in'
			#except Exception, errmsg:
			#	pass

			self.set_header('X-ctxid', ctxid)
			self.set_header('Location', location or '/')

		self.ctxt['msg'] = msg
		self.ctxt['errmsg'] = errmsg


				
						
	@View.add_matcher(r'^/auth/logout/$')
	def logout(self, msg='', location=None, **kwargs):
		self.template = '/pages/auth.login'
		self.title = 'Logout'
		msg = ''
		errmsg = ''
	
		location = location or g.EMEN2WEBROOT		
		if 'auth' in location or not location:
			location = g.EMEN2WEBROOT		
		
		self.set_context_item('location', location)		
		try:
			self.db.logout()
			msg = 'Successfully logged out'
		except Exception, errmsg:
			pass
			
		self.set_header('Location', location or '/')
		self.set_header('X-ctxid', '')
		
		self.ctxt['msg'] = msg
		self.ctxt['errmsg'] = errmsg



	@View.add_matcher(r'^/auth/password/change/$')
	def setpassword(self, location=None, **kwargs):
		self.template = '/pages/auth.password.change'
		self.title = "Password Change"
		# location = location or g.EMEN2WEBROOT		
		# if 'auth' in location or not location:
		# 	location = g.EMEN2WEBROOT			
		self.ctxt['location'] = location
		
		username = kwargs.pop("username",None) or self.db.checkcontext()[0]
		opw = kwargs.pop("opw",None)
		on1 = kwargs.pop("on1",None)
		on2 = kwargs.pop("on2",None)

		msg = ''
		errmsg = ''

		self.ctxt['username'] = username
		
		if not on1 and not on2:
			pass

		elif on1 != on2:					
			errmsg = "New passwords did not match"
			
		else:
			try:
				self.db.setpassword(opw, on1, username=username)
				msg = "Password changed successfully"
			except Exception, errmsg:
				pass
				
		self.ctxt['msg'] = msg
		self.ctxt['errmsg'] = errmsg
			
			
			
	@View.add_matcher(r'^/auth/password/reset/$', r'^/auth/password/reset/(?P<email>.+)/(?P<secret>\w+)/$')
	def resetpassword(self, location=None, email=None, secret=None, newpassword=None, **kwargs):
		self.template = '/pages/auth.password.reset'
		self.title = "Reset Password"
		self.ctxt['email'] = email
		self.ctxt['secret'] = secret
		self.ctxt['newpassword'] = ''
		self.ctxt['location'] = location
		msg = ''
		errmsg = ''
		
		if email:
			if secret and newpassword:
				try:
					username = self.db.setpassword(oldpassword=None, newpassword=newpassword, secret=secret, username=email)
					msg = 'The password for your account, %s, has been changed'%username
				except Exception, errmsg:
					pass
			
			elif secret and not newpassword:
				# errmsg = "No new password given..."				
				pass
				
			else:
				try:					
					self.db.resetpassword(email)
					msg = 'Instructions for resetting your password have been sent to %s'%email
				except Exception, errmsg:
					pass


		self.ctxt['msg'] = msg
		self.ctxt['errmsg'] = errmsg



	@View.add_matcher(r'^/auth/email/change/$')
	def setemail(self, location=None, **kwargs):
		self.template = '/pages/auth.email.change'
		self.title = "Change Email"
		self.ctxt['location'] = location
		
		username = kwargs.get("username") or self.db.checkcontext()[0]
		opw = kwargs.get('opw', None)
		email = kwargs.get('email', None)
		
		self.ctxt['email'] = email
		msg = ''
		errmsg = ''
		
		if email:
			try:
				ret = self.db.setemail(email, password=opw, username=username)
				msg = 'A verification email has been sent to %s'%email
			except Exception, errmsg:
				pass

		self.ctxt['msg'] = msg
		self.ctxt['errmsg'] = errmsg



	@View.add_matcher(r'^/auth/email/verify/(?P<email>.+)/(?P<secret>\w+)/$')
	def verifyemail(self, location=None, email=None, secret=None, **kwargs):
		self.template = '/pages/auth.email.verify'
		self.title = "Verify Email"
		msg = ''
		errmsg = ''
		
		if email and secret:
			try:
				ret = self.db.setemail(email, secret=secret)
				msg = "The email address for your account has been changed to %s"%ret
			except Exception, errmsg:
				pass
		
		self.ctxt['msg'] = msg
		self.ctxt['errmsg'] = errmsg
		

			
__version__ = "$Revision: 1.30 $".split(":")[1][:-1].strip()
