# $Id: map.py,v 1.65 2011/03/08 12:29:02 irees Exp $
import copy
import collections

import emen2.db.config
g = emen2.db.config.g()
from emen2.web.view import View



def dfs(root, tree, recurse=1):
	def inner(stack, children, depth=0):
		if recurse > 0 and depth >= recurse:
			return
		
		for child in children:
			newchildren = tree.get(child, set())
			stack |= newchildren
			inner(stack, newchildren, depth=depth+1)
				
	stack = set()
	inner(stack, tree.get(None))
	return stack
	
	

@View.register
class Map(View):
	
	@View.add_matcher(r'^/sitemap/$', r'^/sitemap/(?P<root>\w+)/$')
	def sitemap(self, root=0, *args, **kwargs):
		self.init(root=root, recurse=1, *args, **kwargs)
		self.template = '/pages/sitemap'
		
	
	#@View.add_matcher(r'^/sitemap/$', r'^/sitemap/(?P<root>\w+)/$', r'^/sitemap/(?P<root>\w+)/(?P<action>\w+)/$')
	@View.add_matcher(r'^/map/(?P<keytype>\w+)/(?P<root>\w+)/(?P<mode>\w+)/$')
	def init(self, root=None, recurse=1, keytype="record", action=None, mode="children", expandable=True, **kwargs):
		self.template = '/pages/map'
		self.title = 'Sitemap'
		self.root = root
		self.keytype = keytype
		self.recurse = int(recurse)
		self.mode = mode
		self.expandable = expandable
		
		# Might not be strictly necessary..
		if self.keytype == "record":
			self.root = int(self.root)

		# Expand all nodes. -3 turns into -1...
		if action=="expand" or self.recurse == -1:
			self.recurse = -3
		
		parents = set()
						
		# add 1 to recurse to get enough info to draw the next level
		if self.mode == "children":
			self.tree = self.db.getchildtree(self.root, recurse=self.recurse+2, keytype=self.keytype)
			# get one level of parents as well..
			parents = self.db.getparents(self.root, keytype=self.keytype) 
		else:
			self.tree = self.db.getparenttree(self.root, recurse=self.recurse+2, keytype=self.keytype)
		
		# connect the root to "None" to simplify drawing..
		self.tree[None] = [self.root]
		
		stack = dfs(self.root, self.tree, recurse=self.recurse)
		stack.add(self.root)
		stack |= parents
		
		recnames = {}
		if self.keytype == "record":
			recnames.update(self.db.renderview(stack, viewtype="recname"))
		else:
			for k in stack:
				recnames[k]=k

		self.ctxt['mode'] = self.mode
		self.ctxt['root'] = self.root
		self.ctxt['tree'] = self.tree
		self.ctxt['recurse'] = self.recurse
		self.ctxt['recnames'] = recnames
		self.ctxt['keytype'] = keytype
		self.ctxt['parents'] = parents
		self.ctxt['expandable'] = expandable
		

# @View.register
# class MapOld(View):
# 
# 	@View.add_matcher(r'^/map/(?P<keytype>\w+)/(?P<root>\w+)/(?P<mode>\w+)/$')
# 	def init(self, keytype="record", root=None, mode="parents", rctx=0, recurse=-1, sparse=1, collapsedrectypes=None, expand=None, maxnodes=100000, showroot=1):
# 
# 		self.template = '/pages/map'
# 		self.keytype = keytype
# 		self.root = root
# 		self.mode = mode
# 		self.rctx = rctx
# 		self.showroot = int(showroot)
# 		self.sparse = sparse
# 		self.maxnodes = maxnodes
# 		self.recnames = {}
# 
# 
# 		self.recurse = int(recurse)
# 		self.expand = map(int, expand or [])
# 		self.collapsed = collections.defaultdict(set)
# 		self.collapsedchildren = {}
# 		self.collapsedrectypes = collapsedrectypes or []
# 		if not self.collapsedrectypes and self.mode in ["both", "children"]:
# 			self.collapsedrectypes = ["scan", "ccd", "ccd_jadas", "micrograph", "grid", "stackimage"]
# 
# 		if self.keytype == "record":
# 			self.root = int(self.root)
# 
# 		# try:
# 		if self.mode == "both":
# 			table, img = self.build_both()
# 		else:
# 			table, img = self.build(mode=self.mode)
# 
# 		# except Exception, inst:
# 		# 	self.template='/errors/error_mapoverride'
# 		# 	self.ctxt["confirm"] = 0
# 		# 	if isinstance(inst,ValueError):
# 		# 		self.ctxt["confirm"] = 1
# 		# 	self.ctxt["mode"] = self.mode
# 		# 	self.ctxt["keytype"] = self.keytype
# 		# 	self.ctxt["root"] = self.root
# 		# 	self.ctxt["errmsg"] = inst
# 		# 	return
# 
# 
# 		self.ctxt["root"] = self.root
# 		self.ctxt["table"] = table
# 		self.ctxt["img"] = img
# 		self.ctxt["recnames"] = self.recnames
# 		self.ctxt["keytype"] = self.keytype
# 		self.ctxt["mode"] = self.mode
# 		self.ctxt["collapsed"] = self.collapsed
# 
# 
# 	def build_both(self):
# 		table1, img1 = self.build(mode="parents")
# 		table2, img2 = self.build(mode="children")
# 
# 		y1, x1 = len(table1), len(table1[0])
# 		y2, x2 = len(table2), len(table2[0])
# 
# 		# Pad length
# 		if x1 < 2:
# 			for y in range(y1):
# 				table1[y].insert(0, None)
# 				img1[y].insert(0, '')
# 			x1 = len(table1[0])
# 		
# 		if x2 < 2:
# 			for y in range(y2):
# 				table2[y].append(None)
# 				img2[y].append('')
# 			x2 = len(table2[0])
# 
# 		# Pad height
# 		h = y1
# 		if y2 > h:
# 			h = y2
# 
# 		for y in range(h-y1):
# 			table1.append([None]*x1)
# 			img1.append([""]*x1)
# 
# 		for y in range(h-y2):
# 			table2.append([None]*x2)
# 			img2.append([""]*x2)
# 
# 		for y in range(h):
# 			table1[y].extend(table2[y][1:])
# 			img1[y].extend(img2[y][:-1])
# 
# 
# 		return table1, img1
# 
# 
# 
# 	def bfs(self, root, tree):
# 		stack = copy.copy(tree.get(root, set()))
# 		seen = set([root])
# 		seen |= stack
# 		while stack:
# 			cur = stack.pop()
# 			children = tree.get(cur, set())
# 			stack |= children
# 			seen |= children
# 		return seen
# 
# 		
# 
# 
# 	def build(self, mode):
# 		
# 		if mode == "children":
# 			results = self.db.getchildtree(self.root, keytype=self.keytype, recurse=self.recurse)
# 		else:
# 			results = self.db.getparenttree(self.root, keytype=self.keytype, recurse=self.recurse)
# 		
# 
# 		if len(results) > self.maxnodes and self.maxnodes:
# 			raise ValueError, """There are %s records in this map, which exceeds the safety limit of %s; are you sure you want to render this? It will likely crash your browser."""%(len(results),self.maxnodes)
# 
# 
# 		r = [self.root]
# 		for i in results.values():
# 			r.extend(i)
# 		r = set(r)
# 
# 		if self.keytype=="record":
# 			rectypes = self.db.groupbyrecorddef(r)
# 			colrecs = []
# 			for i in self.collapsedrectypes:
# 				colrecs.extend(rectypes.get(i,[]))
# 			colrecs = set(colrecs)
# 			uncollapsed = r - colrecs
# 			self.recnames.update(self.db.renderview(uncollapsed, viewtype="recname"))
# 
# 		else:
# 			if self.keytype=="paramdef":
# 				items = self.db.getparamdef(r)
# 			else:
# 				items = self.db.getrecorddef(r)
# 
# 			for i in items:
# 				self.recnames[i.name] = i.desc_short or i.name # "%s<br />%s"%(i.name,i.desc_short)
# 
# 
# 		depth = self.finddepth(self.root, results)
# 
# 		img   = [[""]*(depth+1)]
# 		table = [[None]*(depth+1)]
# 		table[0][0] = self.root
# 
# 
# 		y = 0
# 		x = 0
# 		while y < len(table):
# 	
# 			x = 0
# 			while x < depth:
# 
# 				parent = table[y][x]
# 				children = copy.copy(results.get(parent,set()))
# 				
# 				#print "parent, self.expand:", parent, self.expand
# 				if parent not in self.expand and parent != None and children:
# 					self.recnames[parent] = "%s [+%s]"%(self.recnames.get(parent), len(children))
# 					results[parent] = set()
# 					children = set()
# 
# 
# 				# Sort by recname...
# 				for yoffset,child in enumerate(sorted(children, key=lambda x:self.recnames.get(x,'').lower())):
# 					if yoffset == 0:
# 						table[y][x+1] = child
# 						
# 						
# 					if yoffset > 0:
# 						table.insert(y+yoffset,[None]*(depth+1))
# 						table[y+yoffset][x+1] = child
# 
# 		
# 				if len(children) > 0:
# 					img[y][x]="next"
# 					
# 				if len(children) > 1:
# 					img[y][x]="branch_next"
# 					for yoffset in range(len(children))[1:]:
# 						img.insert(y+yoffset,[""]*(depth+1))
# 						img[y+yoffset][x] = "branch_both"
# 					img[y+yoffset][x] = "branch_up"
# 
# 				# if mode != "children":
# 				# 	print table
# 		
# 				x+=1		
# 			y+=1
# 			
# 
# 		# add spacers; paints down columns looking for gaps
# 		paint=0
# 		for x in range(len(img[0])):
# 			for y in range(len(img)):
# 				if img[y][x]:
# 					paint = 0
# 				if paint:
# 					img[y][x]="branch_spacer"		
# 				if img[y][x] in ["branch_next","branch_both"]:
# 					paint = 1
# 
# 
# 		if mode == "parents":
# 			for i in table:
# 				i.reverse()
# 			for i in img:
# 				i.reverse()
# 				
# 		if mode == "children":
# 			for y in range(len(img)):
# 				for x in range(len(img[0])):
# 					img[y][x] = img[y][x]+"_reverse"
# 
# 		return table, img
# 
# 
# 
# 	def finddepth(self, root, graph):
# 
# 		depth=0
# 
# 		v=[]
# 		for i in graph.values():
# 			v.extend(i)
# 		v=set(v)
# 
# 		f1=graph.get(root,set())
# 		while f1 & v:
# 
# 			#if depth > self.maxrecurse:
# 			#	raise Exception,"Recursion depth exceeded (circular relationship?)"
# 			f2=set()
# 			for i in f1:
# 				f2 |= graph.get(i,set())
# 			f1=f2
# 			depth+=1
# 
# 		return depth



__version__ = "$Revision: 1.65 $".split(":")[1][:-1].strip()
