# $Id: query.py,v 1.63 2011/03/07 09:45:30 irees Exp $
import os
import operator
import copy
import urllib

# Standard View imports
import emen2.util.listops
import emen2.db.config
g = emen2.db.config.g()
from emen2.web.view import View

cmp_order = ["==", "!=", ".contains.", ">=", "<=", ">", "<", ".any.", '.empty.', '.recid.', '.rectype.']


def convert_bytes(bytes):
	bytes = float(bytes)
	if bytes >= 1099511627776:
		terabytes = bytes / 1099511627776
		size = '%.2f TB' % terabytes
	elif bytes >= 1073741824:
		gigabytes = bytes / 1073741824
		size = '%.2f GB' % gigabytes
	elif bytes >= 1048576:
		megabytes = bytes / 1048576
		size = '%.2f MB' % megabytes
	elif bytes >= 1024:
		kilobytes = bytes / 1024
		size = '%.2f KB' % kilobytes
	else:
		size = '%.2f bytes' % bytes
	return size



def query_to_path(q):
	pass



def path_to_query(path, **kwargs):
	q = {}
	q['c'] = []
	
	if path == None:
		path = ''

	# print path	
	c = path.split("/")
	for constraint in c:
		# cmp_order matches in an order so >= will be tested before >, to avoid things like ['param', '>', '=value]
		match = []
		for comp in cmp_order:
			if match:
				continue
			p = urllib.unquote(constraint).partition(comp)
			if p[1]:
				match = [urllib.unquote(p[0]), urllib.unquote(p[1]).replace('.',''), urllib.unquote(p[2])]
		if match: q['c'].append(match)
							
	for k,v in kwargs.items():
		q[urllib.unquote(k)] = urllib.unquote(v)
	
	return q		
	
	
	
class TooManyFiles(Exception):
	pass


@View.register
class Query(View):
	
	def init(self, path, *args, **kwargs):
		self.template = '/pages/query'
		self.title = "Query"

		q = kwargs.pop('q', {})
		if q:
			newq = {}
			for k,v in q.items():
				newq[str(k)] = v
			self.q = newq
			return
			
		self.q = path_to_query(path, **kwargs)

		
	
	@View.add_matcher(r'^/query/$', r'^/query/(?P<path>.*)/$')
	def main(self, path=None, notify=None, *args, **kwargs):
		self.init(path, *args, **kwargs)
		self.q['count'] = 100
		self.ctxt['q'] = self.db.querytable(**self.q)

		

	@View.add_matcher(r'^/query/(?P<path>.*)/edit/$')
	def edit(self, path=None, *args, **kwargs):
		self.init(path, *args, **kwargs)
		self.template = '/pages/query.edit'
		self.ctxt['q'] = self.q

			
		
	# /download/ can't be in the path because of a emen2resource.getchild issue	
	@View.add_matcher(r'^/query/(?P<path>.*)/files/$')
	def download(self, path=None, confirm=False, *args, **kwargs):
		self.init(path, *args, **kwargs)
		self.template = '/pages/query.files'
		self.ctxt['q'] = self.q
		bdos = self.db.getbinary(q=self.q)
		
		for bdo in bdos:
			if bdo.get('filesize') == None:
				if os.access(bdo.get('filepath'), os.F_OK):
					bdo.filesize = os.stat(bdo.get('filepath')).st_size					
			
		if len(bdos) > 10000:
			raise TooManyFiles, "More than 10,000 files returned. Are you sure you want to continue?"
			
		filesize = sum([(bdo.get('filesize') or 0) for bdo in bdos])
		
		users = set([bdo.get('creator') for bdo in bdos])
		users = self.db.getuser(users)
		self.ctxt['users'] = emen2.util.listops.dictbykey(users, 'username')
		
		self.ctxt['rendered'] = self.db.renderview(set([bdo.get('recid') for bdo in bdos]), viewtype="recname")
		
		self.ctxt['filesize'] = convert_bytes(filesize)
		self.ctxt['bdos'] = bdos
		

		
	@View.add_matcher(r'^/query/(?P<path>.*)/goupby/$')
	def groupby(self, notify=None, path=None, *args, **kwargs):
		self.template = '/pages/query.groupby'
				
		# if recid == None:
		# 	recid = g.BOOKMARKS.get("GROUPS", 0)
		# recs = self.db.getchildren(recid, recurse=3, rectype="project")
		# recs = self.db.getrecord(recs)			
		if path == None:
			path = "rectype==project"

		self.q = path_to_query(path, **kwargs)
		self.q = self.db.query(**self.q)
		recids = self.q.get('recids', [])
		recs = self.db.getrecord(recids)
		
		self.title = "Projects"
		
		sortitems = collections.defaultdict(set)
		sortkey = kwargs.get('sortkey', 'project_status')
		for rec in recs:
			try:
				sortitems[rec.get(sortkey)].add(rec.recid)
			except:
				sortitems[",".join(rec.get(sortkey))].add(rec.recid)

		

		children = self.db.getchildren([rec.recid for rec in recs], recurse=-1)
		grouped = self.db.groupbyrecorddef(emen2.util.listops.flatten(children))
		grouped_inverse = emen2.util.listops.invert(grouped)
		
		mostrecents = set()

		details = {}
		for rec in recs:
			print rec.recid
			d = {}
			c = children.get(rec.recid)
			d['grouped'] = collections.defaultdict(set)
			for c1 in c:
				d['grouped'][grouped_inverse.get(c1)].add(c1)

			d['mostrecent'] = {}
			for k,v in d['grouped'].items():
				mr = sorted(v)[-1]
				d['mostrecent'][k] = mr
				mostrecents.add(mr)
			details[rec.recid] = d

		
		mostrecents = self.db.getrecord(mostrecents)
		recs.extend(mostrecents)
		rendered = self.db.renderview(recs, viewtype="recname")
		self.ctxt['recsdict'] = emen2.util.listops.dictbykey(recs, 'recid')

						
		self.ctxt['details'] = details	
		
				
		pages = {
			'classname':'main',
			'labels': {}
		}

		pages['order'] = sortitems.keys()
		for k,v in sortitems.items():
			pages['labels'][k]='%s (%s)'%(k, len(v))
		
		pages = emen2.web.markuputils.HTMLTab(pages)
		self.ctxt['pages'] = pages

		self.ctxt["rendered"] = rendered
		self.ctxt["sortitems"] = sortitems
						
		
# 
# 
# class QueryOld(View):
# 	__metaclass__ = View.register_view
# 	__matcher__ = r'^/queryold/$'
# 
# 	def init(self, username=None, pw=None, query=None, **kwargs):
# 		self.title = "Query"
# 		self.template = "/pages/query"
# 
# 		# ian: will be replaced by new user/record/param chooser
# 		self._getdefs()
# 
# 		###### Process query arguments from forms ###
# 		q = kwargs.get('q')
# 		plot = {} # self._processplot(kwargs)
# 
# 		subset = kwargs.get('subset','').strip()
# 		if subset:
# 			subset = set(map(int, subset.split(",")))
# 		else:
# 			subset = None
# 
# 		boolmode = kwargs.get("boolmode", "AND")
# 
# 
# 		constraints = []
# 		params = kwargs.get("params",[])
# 		paramcompare = kwargs.get("paramcompare",[])
# 		paramvalue = kwargs.get("paramvalue",[])
# 		
# 		# print params
# 		# print paramcompare
# 		# print paramvalue
# 		
# 		for i,p in enumerate(params):
# 			constraints.append([p,paramcompare[i],paramvalue[i]])
# 
# 		qp = {}
# 		qp["q"] = q
# 		qp["constraints"] = constraints
# 		qp["boolmode"] = boolmode
# 		qp["subset"] = subset
# 		qp.update(plot)
# 
# 		# print "Query input"
# 		# print qp
# 
# 		if constraints:
# 			qp = self.db.plot(**qp)
# 
# 		# print "Query results"
# 		# print qp
# 		
# 		#################
# 		# Render tables for results
# 		#################
# 
# 		grouprectype = qp.get('groups', {}).get('rectype', {})
# 		if not grouprectype:
# 			grouprectype = self.db.groupbyrecorddef(qp.get('recids', set()))
# 						
# 
# 		tables = {}
# 		for group, recids in grouprectype.items():
# 			qp2 = copy.deepcopy(qp)
# 			qp2["rectype"] = group
# 			t = emen2.web.views.recordtable.Table(rectype=group, recids=recids, mode="query", query=qp2, db=self.db)
# 			tables[group] = unicode(t.get_data())
# 
# 
# 		self.ctxt["qp"] = qp
# 		self.ctxt["stats"] = {}
# 		self.ctxt["tables"] = tables
# 		self.ctxt["comparators"] = comparators
# 
# 
# 
# 	def _getdefs(self):
# 		pdn = self.db.getparamdefnames()
# 		rdn = self.db.getrecorddefnames()
# 		usernames = self.db.getusernames()
# 		if usernames == None:
# 			usernames = []
# 
# 		userdisplaynames = {}
# 		for user in self.db.getuser(usernames):
# 			userdisplaynames[user.username] = user.displayname
# 
# 		usernames = [i[0] for i in sorted(userdisplaynames.items(), key=operator.itemgetter(1))]
# 
# 		paramdefs = filter(lambda x:x.indexed, self.db.getparamdef(pdn))
# 		paramdef_desc = dict([[v.name,v.desc_short] for v in paramdefs])
# 
# 		paramsK={}
# 		for i in rdn:
# 			try:
# 				paramsK[i] = sorted(self.db.getrecorddef(i).paramsK)
# 			except:
# 				pass
# 
# 		# basic statistics
# 		rdncount = {}
# 		for i in rdn:
# 			rdncount[i] = len(self.db.getindexbyrecorddef(i))
# 		pdncount = {}
# 
# 		self.ctxt["pdn"] = pdn
# 		self.ctxt["rdn"] = rdn
# 		self.ctxt["usernames"] = usernames
# 		self.ctxt["userdisplaynames"] = userdisplaynames
# 		self.ctxt["rdncount"] = rdncount
# 		self.ctxt["pdncount"] = pdncount
# 		self.ctxt["paramsK"] = paramsK
# 		self.ctxt["paramdef_desc"] = paramdef_desc
# 		self.ctxt["paramdefs"] = paramdefs
# 
# 
# 
# 	def _processplot(self, kwargs):
# 		"""We need to transform the html form input slightly to get the plot parameters"""
# 
# 		plotkwargs = kwargs.get('plot', {})
# 		plot = {}
# 		groupshow = kwargs.get('groupshow', {})
# 		groupcolor = kwargs.get('groupcolor', {})
# 		formats = kwargs.get('formats', {})
# 
# 		# copy only arguments we need, just in case..
# 		if plotkwargs:
# 			for i in ["xparam", "yparam", "groupby", "grouptype", "cutoff", "xmin", "xmax", "ymin", "ymax", "width"]:
# 				plot[i] = plotkwargs.get(i)
# 			if plot.get("xparam") and plot.get("yparam"):
# 				pass
# 			else:
# 				# print "Insufficient args for plotting..."
# 				plot = {}
# 
# 		plot["groupshow"] = groupshow.keys() or None
# 		plot["formats"] = formats.keys() or None
# 		plot["groupcolor"] = groupcolor or None
# 
# 		return plot
# 
# 
# 
# 
# def median(numbers):
#    "Return the median of the list of numbers."
#    # Sort the list and take the middle element.
#    n = len(numbers)
#    copy = numbers[:] # So that "numbers" keeps its original order
#    copy.sort()
#    if n & 1:         # There is an odd number of elements
#       return copy[n // 2]
#    else:
#       return (copy[n // 2 - 1] + copy[n // 2]) / 2



# Generate stats..
#
# # if rectype:
# if False:
# 	stats = {"mean":{},"median":{},"n":{},"min":{},"max":{}}
# 	recdef = self.db.getrecorddef(rectype)
# 	p = db.getparamdef(recdef.paramsK)
#
#
# 	stat_excludezero=int(kwargs.get("stat_excludezero",0))
#
# 	for k,v in p.items(): #{}.items():
# 		try:
# 			if v.property != None or v.vartype in ["float","int"]:
# 				z = self.db.getindexdictbyvalue(k,None,subset=queryrecids)
# 				sums=[]
# 				for i in z.values():
# 					if i!=None and not (i==0 and stat_excludezero):
# 						try:
# 							sums.append(float(i))
# 						except:
# 							pass
# 				sums.sort()
# 				if len(sums) > 0:
# 					stats["mean"][k]=sum(sums)/len(sums)
# 					stats["median"][k]=median(sums)
# 					stats["n"][k]=len(sums)
# 					stats["min"][k]=sums[0]
# 					stats["max"][k]=sums[-1]
# 				else:
# 					stats["mean"][k]=""
# 					stats["median"][k]=""
# 					stats["n"][k]=0
# 					stats["min"][k]=""
# 					stats["max"][k]=""
#
# 		except:
# 			pass
#
#
# 	stat_roundfloats=int(kwargs.get("stat_roundfloats",0))
# 	if stat_roundfloats:
# 		self.ctxt["fstr"]="%0.3f"
# 	else:
# 		self.ctxt["fstr"]="%s"
#
# 	self.ctxt["stat_roundfloats"]=stat_roundfloats
# 	self.ctxt["stat_excludezero"]=stat_excludezero
# 	self.ctxt["paramdefs"]=p
# 	self.ctxt["stats"]=stats



# class QueryNew2(View):
# 		__metaclass__ = View.register_view
# 		__matcher__ = [r'^/query2/(?P<query>.*)/$',
# 														r'^/query2/$']
#
#
# 		def query_new2(self):
# 				self._query = query.split('|')
# 				self._query = [ item.strip().split(" ") for item in self._query ]
# 				self._query = [ (item[0], [ x.strip() for x in listops.get(item, 1, '').split(",") ]) for item in self._query ]
#
# 				self._dbquery = q.DBQuery(db)
#
# 				predicates = dict( getrecord = q.GetRecord(),
# 													getchildren = q.GetChildren(),
# 													# getparents = q.GetParents,
# 													filter = q.FilterByValue )
# 				special = set(['filter'])
# 				tmp = []
# 				for pred, args in	 self._query:
# 						args = filter(bool, args)
# 						obj = predicates.get(pred)
# 						if pred in special:
# 							obj = obj(*args)
# 						tmp.append(obj)
#
# 				self._query = tmp
# 				self.get_data = self.proc
#
# 		def form(self):
# 				form = """\
# 				<html>
# 						<head><title>Form</title>
# 						<body>
# 								 <form action=".">
# 										 <textarea name="query" style="width:100%" rows="30"></textarea>
# 										 <input type="submit"></input>
# 								 </form>
# 						</body>
# 				</html>
# 				"""
# 				return form
#
# 		def proc(self):
# 				self._dbquery.reset()
# 				for pred in self._query:
# 						self._dbquery <<= pred
# 				return "<pre>" + str.join("", [str(x) for x in self._dbquery.result]) + "</pre>"
#
__version__ = "$Revision: 1.63 $".split(":")[1][:-1].strip()
